#include <string.h>
#include "arg.h"
#include "log.h"
#include "dep.h"
#include "generator.h"

/* for strue/sfalse */
#include "db.h"

int notree = 0, want_debug = 0, want_optim = 1;
char *notree_reason = NULL;

static void help(const char *topic)
{
	printf("\nconfigure - configures libporty and c99tree before compilation\n");
	printf("\nusage: ./configure [--key1[=val]] [--key2[=val]] [...]\n");
	printf("\nPossible key=value pairs:\n");
	printf("  --prefix=path       changes installation prefix (default: /usr/local)\n");
	printf("  --without-c99tree   do not compile c99tree, only libporty\n");
	printf("  --debug             compile for debug (normally -O0 -g ...)\n");
	printf("  --symbols           compile with debug symbols (-g)\n");
	printf("\n");
}


/* Runs when a custom command line argument is found
 returns true if no furhter argument processing should be done */
int hook_custom_arg(const char *key, const char *value)
{
	if (strcmp(key, "help") == 0) {
		help(value);
		exit(0);
	}
	if (strcmp(key, "without-c99tree") == 0) {
		notree = 2;
		notree_reason = "it was disabled from the command line (--without-c99tree)";
		return 1;
	}
	if (strcmp(key, "prefix") == 0) {
		put("/local/prefix", value);
		return 1;
	}
	if (strcmp(key, "debug") == 0) {
		want_debug = 1;
		want_optim = 0;
		return 1;
	}
	if (strcmp(key, "symbols") == 0) {
		want_debug = 1;
		return 1;
	}
	return 0;
}


/* Runs before anything else */
int hook_preinit()
{
	return 0;
}

/* Runs after initialization */
int hook_postinit()
{
	report("Configuring libporty.\n");
	logprintf(0, "Configuring libporty.\n");
	db_mkdir("/local");
	put("/local/prefix", "/usr/local");
	put("/local/target/cflags", "");
	put("/local/target/ldflags", "");
	return 0;
}

/* Runs after all arguments are read and parsed */
int hook_postarg()
{
	return 0;
}

/* Runs when things should be detected for the host system */
int hook_detect_host()
{
	require("fstools/ar", 1, 1);
	require("fstools/ranlib", 1, 1);
	require("fstools/mkdir", 1, 1);
	return 0;
}

#include "hooks_net.c"

static void spappend(const char *path, const char *val)
{
	append(path, " ");
	append(path, val);
}

/* Runs when things should be detected for the target system */
int hook_detect_target()
{
	const char *pic;

	if (want_debug) {
		const char *ansi, *ped, *wall;
		require("cc/argstd/*", 0, 0);
		ansi = get("/target/cc/argstd/ansi");
		ped = get("/target/cc/argstd/pedantic");
		wall = get("/target/cc/argstd/Wall");

		if (ansi != NULL) spappend("/local/target/cflags", ansi);
		if (ped != NULL)  spappend("/local/target/cflags", ped);
		if (wall != NULL) spappend("/local/target/cflags", wall);
		spappend("/local/target/cflags", "-Dinline=");
	}

	if (want_optim)
		append("/local/target/cflags", "-O2");

	require("cc/fpic",  0, 0);
	pic = get("/target/cc/fpic");
	if (pic != NULL)  spappend("/local/target/cflags", pic);

	if (require("libs/parser/expat/ldflags", 1, 0)) {
		notree = 1;
		notree_reason = "of missing libexpat";
	}

	if (!notree && require("parsgen/flex", 1, 0)) {
		notree = 1;
		notree_reason = "of missing flex";
	}

	if (!notree && require("parsgen/bison", 1, 0)) {
		notree = 1;
		notree_reason = "of missing bison";
	}

	libporty_net_detect_target();

	/* detect math lib corner cases for math_wrap.c */
	require("libs/math/cc/log/*", 1, 0);

	if (notree)
		put("local/c99tree/enabled", sfalse);
	else
		put("local/c99tree/enabled", strue);

	return 0;
}

/* Runs after detection hooks, should generate the output (Makefiles, etc.) */
int hook_generate()
{
	printf("Generating Makefiles for src/...\n");
	generate("../src/Makefile.in", "../src/Makefile");
	printf("Generating Makefiles for c99tree...\n");
	generate("../src/porty/c99tree-cc/Makefile.config.in", "../src/porty/c99tree-cc/Makefile.config");
	printf("Generating config.h for libporty...\n");
	generate("../src/libporty/config.h.in", "../src/libporty/config.h");
	generate("../src/libporty/Makefile.config.in", "../src/libporty/Makefile.config");

	generate("../src/libporty/net/os_includes.h.in", "../src/libporty/net/os_includes.h");
	generate("../src/libporty/net_resolver/resolver_inc.h.in", "../src/libporty/net_resolver/resolver_inc.h");
	generate("../src/libporty/host/fs_includes.h.in", "../src/libporty/host/fs_includes.h");
	generate("../src/libporty/host/time_includes.h.in", "../src/libporty/host/time_includes.h");
	generate("../src/libporty/host/types.h.in", "../src/libporty/host/types.h");
	generate("../src/libporty/net_async/Makefile.in", "../src/libporty/net_async/Makefile");
	generate("../src/regression/libporty/Makefile.in", "../src/regression/libporty/Makefile");


	printf("\n\n=== Result ===\n");
	if (notree) {
		printf("libporty is configured.\n");
		printf("Can't configure (and build) c99tree because %s\n", notree_reason);
		if (notree == 1)
			printf("In case c99tree and porty is needed, please install the missing software and rerun configure\n");
		printf("(This is not an error if you need only libporty and don't need c99tree)\n");
	}
	else
		printf("libporty and c99tree and porty are configured.\n");
	return 0;
}

/* Runs before everything is uninitialized */
void hook_preuninit()
{
}

/* Runs at the very end, when everything is already uninitialized */
void hook_postuninit()
{
}

