/*
    libporty - local socket pairs
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Contact (email and IRC): http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include "local.h"
#include "network.h"

#ifdef P_HAVE_SOCKETPAIR
/* simple implementation with dedicated libc call */
int P_net_local_socket(P_net_socket *sockfd1, P_net_socket *sockfd2, P_net_nonblock_t non_blocking)
{
	int fd[2];
	int ret;

	ret = socketpair(P_AF_LOCAL, SOCK_STREAM, 0, fd);
	if (ret != 0)
		return P_net_err_alloc;

	/* Switch to non-blocking, if needed */
	if (non_blocking & P_net_nonblock_socket) {
		if (((P_net_set_nonblocking(fd[0]) != 0) || (P_net_set_nonblocking(fd[1]) != 0)) && (non_blocking & P_net_nonblock_quit_on_error)) {
			P_net_close(fd[0]);
			P_net_close(fd[1]);
			return P_net_err_nonblock;
		}
	}


	*sockfd1 = fd[0];
	*sockfd2 = fd[1];
	return 0;
}
#else
/* No support from the system - safe bet is using a local tcp connection:
    - pollable
    - can be set nonblock
    - bidirectional
    - stream
    - ... and slow
*/
#include "tcp4.h"
int P_net_local_socket(P_net_socket *sockfd1, P_net_socket *sockfd2, P_net_nonblock_t non_blocking)
{
	int port1, port2, rem_port;
	P_net_socket listen;
	struct P_pollfd fds[1];
	int tries, ret, i;

	for(tries = 0; tries < 64; tries++) {
		port1 = 32786 + rand() % 32767;
		port2 = 32786 + rand() % 32767;
		if (P_tcp4_listen(&listen, "127.0.0.1", port1, P_net_nonblock_full, 1) != 0)
			continue;
		if (P_tcp4_connect(sockfd1, "127.0.0.1", port1, "127.0.0.1", port2, P_net_nonblock_connect) != 0) {
			close(listen);
			continue;
		}

		/* wait for connection */
		fds[0].fd = listen;
		fds[0].events = P_POLLIN;
		for(i = 0; i < 4; i++) {
			P_poll(fds, 1, 1);
			if (fds[0].revents & P_POLLIN) {
				/* incoming connection */
				ret = P_tcp4_accept(sockfd2, listen, 0, NULL, &rem_port, non_blocking);

				if (ret == 0) {
				 	if (rem_port == port2) {
						close(listen);
						if (non_blocking & P_net_nonblock_socket) {
							int r1, r2;
							r1 = P_net_set_nonblocking(*sockfd1) != 0;
							r2 = P_net_set_nonblocking(*sockfd2) != 0;
							/* if we couldn't set both to non-blcking and this is considered fatal, clean up and return error */
							if ((r1 || r2) && (non_blocking & P_net_nonblock_quit_on_error)) {
								P_net_close(*sockfd1);
								P_net_close(*sockfd2);
								return P_net_err_nonblock;
							}
						}
						else {
							/* we need to force sockets back to blocking - non-blocking conncet leaves the socket in non-blocking mode for example */
							P_net_set_blocking(*sockfd1);
							P_net_set_blocking(*sockfd2);
						}
						return 0;
					}
					close(*sockfd2);
				}
			}
		}
		close(listen);
		close(*sockfd1);
	}
	return P_net_err_alloc;
}

#endif
