%option reentrant
%option bison-bridge

%{
/*
 * Copyright (c) 2002 Anders Magnusson. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

 /* This is a modified verison for c99tree; these modifications are
    Copyright (c) 2010-2015 Tibor Palinkas.
    Original file version marking:
    scan.l,v 1.94 2010/08/11 14:08:44 ragge Exp */
#undef YY_NO_INPUT
# define LLINEDEF  __LINE__
#define YY_DECL int c99c_lex(YYSTYPE * yylval_param ,yyscan_t yyscanner, c99tree_cc_t *ctx)
%}


D     [0-9]
L     [a-zA-Z_]
H     [a-fA-F0-9]
E     [Ee][+-]?{D}+
P     [Pp][+-]?{D}+
FS    (f|F|l|L)?i?
IS    (u|U|l|L)*
UL    ({L}|\\u{H}{H}{H}{H}|\\U{H}{H}{H}{H}{H}{H}{H}{H})
NL    [\\][\n]
ANY   (.|{NL})

%{
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "c99tree-cc.h"
#include "scope.h"
#include "cgram.h"


static int is_typename(c99tree_cc_t *ctx, YYSTYPE * yylv);
static void control(c99tree_cc_t *ctx, YYSTYPE * yylv, char *text, int t);
static void pragma(c99tree_cc_t *ctx, YYSTYPE * yylv, char *text);

#define CPP_IDENT 2
#define CPP_LINE 3
#define CPP_HASH 4
#define CPP_TREE 5
#define CPP_BREAK 6

#if defined(FLEX_SCANNER) && YY_FLEX_SUBMINOR_VERSION == 31
/* Hack to avoid unnecessary warnings */
FILE *yyget_in  (void);
FILE *yyget_out  (void);
int yyget_leng  (void);
char *yyget_text  (void);
void yyset_in (FILE *);
void yyset_out (FILE *);
int yyget_debug  (void);
void yyset_debug (int);
int yylex_destroy  (void);
extern int yyget_lineno (void);
extern void yyset_lineno (int);
#endif

static void lus(c99tree_cc_t *ctx, YYSTYPE * yylv)
{
	c99c_loc_cpy(&(yylv->string.loc), ctx->loc_current);
}

/* some of these have to be a macro so that c99c_tt_append will use the
   right __LINE__ */

/* shorthand for loc_update_end for keeping track of location */
#define lue_typ(ctx, yylv, text, toktyp) \
do { \
	c99c_loc_update_end(ctx, text); \
	lus(ctx, yylv); \
	c99c_tt_append(ctx, toktyp, c99c_strclone(text)); \
	yylv->string.loc.twin_parent = ctx->loc_current->twin_parent; \
} while(0)

#define lue(ctx, yylv, text) lue_typ(ctx, yylv, text, T_TOKEN_STR)

#define lue_chr() \
do { \
	yylval->string.str = NULL; \
	c99c_loc_update_end(ctx, yytext);\
	c99c_loc_cpy(&(yylval->string.loc), ctx->loc_current); \
	c99c_tt_append(ctx, T_TOKEN_STR, c99c_strclone(yytext)); \
	yylval->string.loc.twin_parent = ctx->loc_current->twin_parent; \
} while(0)

/* same as lue, but also c99c_strclones yytext into yylval.string and saves start of string */
#define luedup() \
do { \
	yylval->string.str = c99c_strclone(c99c_loc_update_end(ctx, yytext)); \
	lus(ctx, yylval); \
	c99c_tt_append(ctx, T_TOKEN_STR, c99c_strclone(yytext)); \
	yylval->string.loc.twin_parent = ctx->loc_current->twin_parent; \
} while(0)

/* same as lue but don't create a node in the token tree */
static void lue_silent(c99tree_cc_t *ctx, YYSTYPE * yylv, char *text)
{
	c99c_loc_update_end(ctx, text);
	lus(ctx, yylv);
}

#define RETURN_COP_B   lue(ctx, yylval, yytext); lus(ctx, yylval); yylval->string.str = c99c_strclone(yytext);  return(C_CUSTOP_B);
#define RETURN_COP_U   lue(ctx, yylval, yytext); lus(ctx, yylval); yylval->string.str = c99c_strclone(yytext);  return(C_UNOP);

static void read_comment(c99tree_cc_t *ctx, YYSTYPE * yylv, yyscan_t scn);

/* remove the last newline from the token tree - useful in # lines: the last
newline is an artificial one inserted by cpp for the # */
static void kill_last_nl(c99tree_cc_t *ctx)
{
	c99c_node_t *last = ctx->TOKT, *prev = NULL;
	if (last != NULL) {
		/* find the last two nodes of the token tree */
		while(last->next != NULL) {
			prev = last;
			last = last->next;
		}

		if (last != NULL) {
			if (last->name[0] == '\n') {
				if ((prev != NULL) && (last->loc.twin_parent == NULL)) {
					prev->next = NULL;
					C99Nfree(ctx, last);
				}
				else {
					/* if we can't safely remove the node, make it an empty string */
					last->name[0] = '\0';
				}
			}
		}
	}
}

%}

%%
"/*" { read_comment(ctx, yylval, yyscanner); }

"//"(.|{NL})*       {
		lue_silent(ctx, yylval, yytext);
		lus(ctx, yylval);
		c99c_tt_append(ctx, T_COMMENT, c99c_strclone(yytext));
	}

"__func__"       { luedup(); return(C_STRING); }
"__attribute__"  { lue(ctx, yylval, yytext);    return(C_ATTRIBUTE); }
"asm"            { lue(ctx, yylval, yytext);    return(C_ASM); }
"auto"           { luedup(); return(C_CLASS); }
"_Bool"          { luedup(); return (C_TYPE); }
"break"          { luedup(); return(C_BREAK); }
"case"           { luedup(); return(C_CASE); }
"char"           { luedup(); return (C_TYPE); }
"_Complex"       { luedup(); return (C_TYPESPEC); }
"const"          { luedup(); return (C_QUALIFIER); }
"continue"       { luedup(); return(C_CONTINUE); }
"default"        { luedup(); return(C_DEFAULT); }
"do"             { luedup(); return(C_DO); }
"double"         { luedup(); return (C_TYPE); }
"else"           { lue(ctx, yylval, yytext);    return(C_ELSE); }
"enum"           { lue(ctx, yylval, yytext);    return(C_ENUM); }
"extern"         { luedup(); return (C_CLASS); }
"__thread"       { luedup(); return (C_CLASS); }
"float"          { luedup(); return (C_TYPE); }
"for"            { luedup(); return(C_FOR); }
"goto"           { luedup(); return(C_GOTO); }
"if"             { luedup(); return(C_IF); }
"_Imaginary"     { luedup(); return (C_TYPESPEC); }
"inline"         { luedup(); return(C_INLINE); }
"int"            { luedup(); return (C_TYPE); }
"long"           { luedup(); return (C_TYPESPEC); }
"register"       { luedup(); return (C_CLASS); }
"restrict"       { luedup(); return (C_QUALIFIER); }
"return"         { luedup(); return(C_RETURN); }
"short"          { luedup(); return (C_TYPESPEC); }
"signed"         { luedup(); return (C_TYPESPEC); }
"sizeof"         { luedup(); return(C_SIZEOF); }
"typeof"         { luedup(); return(C_TYPEOF); }
"static"         { luedup(); return (C_CLASS); }
"struct"         { luedup(); return(C_STRUCT); }
"switch"         { luedup(); return(C_SWITCH); }
"typedef"        { luedup(); return (C_TYPEDEF); }
"union"          { luedup(); return(C_STRUCT); }
"unsigned"       { luedup(); return (C_TYPESPEC); }
"void"           { luedup(); return (C_TYPE); }
"volatile"       { luedup(); return (C_QUALIFIER); }
"while"          { luedup(); return(C_WHILE); }

{UL}({UL}|{D})*  { luedup(); return is_typename(ctx, yylval);}

0[xX]{H}+{IS}?   { luedup(); return(C_ICON); }
0{D}+{IS}?       { luedup(); return(C_ICON); }
{D}+{IS}?        { luedup(); return(C_ICON); }
L'(\\.|[^\\'])+' { luedup(); return(C_ICON); }
'(\\.|[^\\'])+'  { luedup(); return(C_ICON); }

{D}+{E}{FS}?             { luedup(); return(C_FCON); }
{D}*"."{D}+({E})?{FS}?   { luedup(); return(C_FCON); }
{D}+"."{D}*({E})?{FS}?   { luedup(); return(C_FCON); }
0[xX]{H}*"."{H}+{P}{FS}? { luedup(); return(C_FCON); }
0[xX]{H}+"."{P}{FS}?     { luedup(); return(C_FCON); }
0[xX]{H}+{P}{FS}?        { luedup(); return(C_FCON); }

L?\"(\\.|[^\\"]|{NL})*\"      { /*"*/ luedup(); return C_STRING; }

"..."         { lue(ctx, yylval, yytext); return(C_ELLIPSIS); }
">>="         { luedup(); return(C_ASOP); }
"<<="         { luedup(); return(C_ASOP); }
"+="          { luedup(); return(C_ASOP); }
"-="          { luedup(); return(C_ASOP); }
"*="          { luedup(); return(C_ASOP); }
"/="          { luedup(); return(C_ASOP); }
"%="          { luedup(); return(C_ASOP); }
"&="          { luedup(); return(C_ASOP); }
"^="          { luedup(); return(C_ASOP); }
"|="          { luedup(); return(C_ASOP); }
">>"          { luedup(); return(C_SHIFTOP); }
"<<"          { luedup(); return(C_SHIFTOP); }
"++"          { luedup(); return(C_INCOP); }
"--"          { luedup(); return(C_INCOP); }
"&&"          { luedup(); return(C_ANDAND); }
"||"          { luedup(); return(C_OROR); }
"<="          { luedup(); return(C_RELOP); }
">="          { luedup(); return(C_RELOP); }
"=="          { luedup(); return(C_EQUOP); }
"!="          { luedup(); return(C_EQUOP); }
";"           { lue_chr();    return(';'); }
("{"|"<%")    { lue_chr();    return('{'); }
("}"|"%>")    { lue_chr();    return('}'); }
","           { lue_chr();    return(','); }
":"           { lue_chr();    return(':'); }
"="           { lue_chr();    return('='); }
"("           { lue_chr();    return('('); }
")"           { lue_chr();    return(')'); }
("["|"<:")    { lue_chr();    return('['); }
("]"|":>")    { lue_chr();    return(']'); }
"&"           { lue_chr();    return('&'); }
"->"          { lue(ctx, yylval, yytext); lus(ctx, yylval); yylval->string.str = c99c_strclone("-> binary");  return(C_STROP); }
"."           { lue(ctx, yylval, yytext); lus(ctx, yylval); yylval->string.str = c99c_strclone(". binary");   return(C_STROP); }
"!"           { lue(ctx, yylval, yytext); lus(ctx, yylval); yylval->string.str = c99c_strclone("! unary");    return(C_UNOP); }
"~"           { lue(ctx, yylval, yytext); lus(ctx, yylval); yylval->string.str = c99c_strclone("~ unary");    return(C_UNOP); }
"-"           { lue_chr();    return('-'); }
"+"           { lue_chr();    return('+'); }
"*"           { lue_chr();    return('*'); }
"/"           { luedup(); return(C_DIVOP); }
"%"           { luedup(); return(C_DIVOP); }
"<"           { luedup(); return(C_RELOP); }
">"           { luedup(); return(C_RELOP); }
"^"           { lue_chr();    return('^'); }
"|"           { lue_chr();    return('|'); }
"?"           { lue_chr();    return('?'); }

^#pragma[ \t]{ANY}*[\r]?[\n]        { kill_last_nl(ctx); pragma(ctx, yylval, yytext); return C_PRAGMA; }
^#ident[ \t]{ANY}*[\r]?[\n]         { kill_last_nl(ctx); control(ctx, yylval, yytext, CPP_IDENT); }
^#line[ \t]{ANY}*[\r]?[\n]          { kill_last_nl(ctx); control(ctx, yylval, yytext, CPP_LINE); }
^#tree[ \t]{ANY}*[\r]?[\n]          { kill_last_nl(ctx); control(ctx, yylval, yytext, CPP_TREE); }
^#include[ \t]{ANY}*[\r]?[\n]       { kill_last_nl(ctx); lue_typ(ctx, yylval, yytext, T_TOKEN_PPDIR); }
^#include_next[ \t]{ANY}*[\r]?[\n]  { kill_last_nl(ctx); lue_typ(ctx, yylval, yytext, T_TOKEN_PPDIR); }
^#break[ \t]{ANY}*[\r]?[\n]         { kill_last_nl(ctx); control(ctx, yylval, yytext, CPP_BREAK); }
^#{ANY}*[\r]?[\n]                   { kill_last_nl(ctx); control(ctx, yylval, yytext, CPP_HASH); }

[=][<>@#$%)}\].:,`^|/]        {  RETURN_COP_B; }
[:][<>@#$%)}\].:,`^|/=]       {  RETURN_COP_B; }
[@][<>@#$%)}\].:,`^|/=]       {  RETURN_COP_B; }
[`][<>@#$%)}\].:,`^|/=]       {  RETURN_COP_B; }
[<][>@#$%)}\].:,`^|/]         {  RETURN_COP_B; }
[>][<@#$%)}\].:,`^|/]         {  RETURN_COP_B; }
[-][<@#$%)}\]:,`^|/]          {  RETURN_COP_B; }
[+][<@#$%)}\]:,`^|/]          {  RETURN_COP_B; }

[`#$@]                        {  RETURN_COP_U; }




[ \t\v\f\r]+     { lue(ctx, yylval, yytext); }
"\n"             { lue(ctx, yylval, yytext); }
.                { lue(ctx, yylval, yytext); fprintf(stderr, "Warning: lex ignoring character '%c'\n", *yytext); /* ignore bad characters */ }

%%
static void check_switch_file(c99tree_cc_t *ctx, char *fn)
{
	c99c_loc_switch_file(ctx, fn, fn);
}

int
c99c_wrap (yyscan_t yyscanner )
{
	if (0) yyunput(0, NULL, yyscanner); /* gcc warning on unused yyunput() */
	return(1);
}

static void cpp_break()
{
	/* forced breakpont */
}

#define fn_cleanup(wr, str,  res) \
	do { \
		res = NULL; \
		wr = str; \
		while (*wr && *wr != '\"') \
			wr++; \
		if (*wr == 0) \
			return; \
		if (*wr++ != '\"') \
			goto bad; \
		res = wr; \
		while (*wr && *wr != '\"') \
			wr++; \
		if (*wr != '\"') \
			goto bad; \
		*wr = 0; \
	} while(0)

void
control(c99tree_cc_t *ctx, YYSTYPE * yylv, char *text, int t)
{
	char *wr = text;
	char *eptr, *fn;
	int val;

	lue_silent(ctx, yylv, text);

	wr++;             /* Skip initial '#' */
	switch (t) {

	case CPP_BREAK:
		cpp_break();
		return;

	case CPP_IDENT:
		return;         /* Just skip these for now. */

	case CPP_LINE:
		wr += 4;
		/* FALLTHROUGH */
	case CPP_HASH:
		val = strtol(wr, &eptr, 10);
		if (wr == eptr) /* Illegal string */
			goto bad;
		fn_cleanup(wr, eptr, fn);
		check_switch_file(ctx, fn);
		ctx->loc_current->cpp_file = c99c_loc_cpp_file_lookup(ctx, fn);
		ctx->loc_current->cpp_line = val - 1;
		ctx->loc_current->eline = val - 1;
/*		fprintf(stderr, "#HASH: %d\n", ctx->loc_current->cpp_line);*/
		break;

	case CPP_TREE:
		wr += 5;

		while(isspace(*wr)) wr++;
		if (strncmp(wr, "macro_expand_begin", 18) == 0) {
			char *e, *s = c99c_strclone(wr+23);
			c99c_node_t *n;
			e = s + strlen(s)-1;
			*e = '\0';
			ctx->loc_current->macro_level++;
			n = c99c_tt_append(ctx, T_TOKEN_STR, s);
			if (n != NULL)
				n->loc.soffs = -1;
			if (ctx->loc_current->macro_level == 1) {
				ctx->macro_expand_tt = n;
			}
		}
		else if (strncmp(wr, "macro_expand_end", 16) == 0) {
			ctx->loc_current->macro_level--;
			if (ctx->loc_current->macro_level == 0) {
				ctx->macro_expand_tt = NULL;
			}
		}
		else if (strncmp(wr, "offs", 4) == 0) {
			char *end;
			int b1, b2;
			b1 = strtol(wr+5, &end, 10);
			while(isspace(*end)) end++;
			b2 = strtol(end, &end, 10);

			if ((ctx->macro_expand_tt != NULL) && (ctx->macro_expand_tt->loc.soffs == -1)) {
				ctx->macro_expand_tt->loc.soffs = b1;
				ctx->macro_expand_tt->loc.eoffs = b2;
			}

			fn_cleanup(wr, end, fn);
			check_switch_file(ctx, fn);
			ctx->loc_current->offs = ctx->loc_current->soffs = b1;
			if (b2 >= 0)
				ctx->loc_current->eoffs = b2;
		}
		ctx->loc_current->cpp_line--;
		ctx->loc_current->sline--;
		ctx->loc_current->eline--;
		break;
	}
	return;
bad:
	fprintf(stderr, "%s: illegal control", text);
}

/*
 * got a full pragma line.  Split it up here.
 */
static void pragma(c99tree_cc_t *ctx, YYSTYPE * yylv, char *text)
{
	char *arg;

	lus(ctx, yylv);
	lue(ctx, yylv, text);
	arg = text+7;

	while((*arg == ' ') || (*arg == '\t')) arg++;

	yylv->string.str = c99c_strclone(arg);
}

static int is_typename(c99tree_cc_t *ctx, YYSTYPE * yylv)
{
	const c99c_node_t *n;
	n = c99c_scope_get_typedef(ctx, yylv->string.str);
	if (n != NULL) {
		if (ctx->c_typename != NULL)
			free(ctx->c_typename);
		ctx->c_typename = yylv->string.str;
		yylv->node = C99Nclone(ctx, n, INV_LOC_ALL_BUT_ID);
		c99c_loc_cpy(&yylv->node->loc, ctx->loc_current);
		return C_TYPENAME;
	}
	return C_NAME;
}


static void cappend(c99tree_cc_t *ctx, char c)
{
	if (ctx->comment_used == ctx->comment_alloced) {
		ctx->comment_alloced += 256;
		ctx->comment = realloc(ctx->comment, ctx->comment_alloced);
	}
	ctx->comment[ctx->comment_used] = c;
	ctx->comment_used++;
}


/* read in comment as documented in flex doc */
static void read_comment(c99tree_cc_t *ctx, YYSTYPE * yylv, yyscan_t scn)
{
	int c;
	cappend(ctx, '/');
	cappend(ctx, '*');
  for (;;) {
		while(((c = input(scn)) != '*') && (c != EOF)) cappend(ctx, c);    /* eat up text of comment */
    if (c == '*') {
			cappend(ctx, c);
			while ((c = input(scn)) == '*') cappend(ctx, c);
	    if ( c == '/' ) {
				cappend(ctx, c);
				ctx->comment[ctx->comment_used] = '\0';
				yylv->string.str = ctx->comment;
				c99c_loc_update_end(ctx, yylv->string.str);
				ctx->comment = 0;
				ctx->comment_alloced = 0;
				ctx->comment_used = 0;
				c99c_tt_append(ctx, T_COMMENT, yylv->string.str);
				return;
			}
			else
				cappend(ctx, c);
	  }
    if (c == EOF) {
			fprintf(stderr, "EOF in /* */ comment");
			break;
		}
	}

	/* we never get here */
}
